# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from django.db import models


class Author(models.Model):
    name = models.CharField(max_length=32, unique=True)
    age = models.IntegerField()



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from django.db import connections


def run_query(alias, sql, params=None):
    with connections[alias].cursor() as cursor:
        cursor.execute(sql, params)



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import inspect
import re
from collections import Mapping, Sequence
from functools import wraps
from types import MethodType

import six
from django.conf import settings
from django.core.cache import caches

from .utils import sorted_names


class CacheOp(object):

    def __init__(self, alias, operation, key_or_keys):
        self.alias = alias
        self.operation = operation
        if isinstance(key_or_keys, six.string_types):
            self.key_or_keys = self.clean_key(key_or_keys)
        elif isinstance(key_or_keys, (Mapping, Sequence)):
            self.key_or_keys = sorted(self.clean_key(k) for k in key_or_keys)
        else:
            raise ValueError("key_or_keys must be a string, mapping, or sequence")

    @classmethod
    def clean_key(cls, key):
        """
        Replace things that look like variables with a '#' so tests aren't affected by random variables
        """
        for var_re in cls.VARIABLE_RES:
            key = var_re.sub('#', key)
        return key

    VARIABLE_RES = (
        # Long random hash
        re.compile(r'\b[0-9a-f]{32}\b'),
        # UUIDs
        re.compile(r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}'),
        # Integers
        re.compile(r'\d+'),
    )

    def __eq__(self, other):
        return (
            isinstance(other, CacheOp) and
            self.alias == other.alias and
            self.operation == other.operation and
            self.key_or_keys == other.key_or_keys
        )


class CacheRecorder(object):
    """
    Monkey patches a cache class to call 'callback' on every operation it calls
    """
    def __init__(self, alias, callback):
        self.alias = alias
        self.callback = callback

    def __enter__(self):
        cache = caches[self.alias]

        def call_callback(func):
            alias = self.alias
            callback = self.callback

            @wraps(func)
            def inner(self, *args, **kwargs):
                # Ignore operations from the cache class calling itself

                # Get the self of the parent via stack inspection
                frame = inspect.currentframe()
                try:
                    frame = frame.f_back
                    is_internal_call = frame.f_locals.get('self', None) is self
                finally:
                    # Always delete frame references to help garbage collector
                    del frame

                if not is_internal_call:
                    callback(CacheOp(
                        alias=alias,
                        operation=six.text_type(func.__name__),
                        key_or_keys=args[0],
                    ))

                return func(*args, **kwargs)
            return inner

        self.orig_methods = {name: getattr(cache, name) for name in self.cache_methods}
        for name in self.cache_methods:
            orig_method = self.orig_methods[name]
            setattr(
                cache,
                name,
                MethodType(call_callback(orig_method), cache)
            )

    def __exit__(self, _, __, ___):
        cache = caches[self.alias]
        for name in self.cache_methods:
            setattr(cache, name, self.orig_methods[name])
        del self.orig_methods

    cache_methods = (
        'add',
        'decr',
        'delete',
        'delete_many',
        'get',
        'get_many',
        'incr',
        'set',
        'set_many',
    )


class AllCacheRecorder(object):
    """
    Launches CacheRecorders on all the active caches
    """
    def __init__(self, callback):
        self.callback = callback

    def __enter__(self):
        self.recorders = []
        for name in sorted_names(settings.CACHES.keys()):
            recorder = CacheRecorder(name, self.callback)
            recorder.__enter__()
            self.recorders.append(recorder)

    def __exit__(self, type_, value, traceback):
        for recorder in reversed(self.recorders):
            recorder.__exit__(type_, value, traceback)
        self.recorders = []



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import shutil
from tempfile import mkdtemp

import pytest
import six
import yaml
from django.test import SimpleTestCase

from django_perf_rec.yaml import KVFile


class KVFileTests(SimpleTestCase):

    def setUp(self):
        super(KVFileTests, self).setUp()
        KVFile._clear_load_cache()
        self.temp_dir = mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_dir)
        super(KVFileTests, self).tearDown()

    def test_load_no_permissions(self):
        with pytest.raises(IOError):
            KVFile('/')

    def test_load_non_existent_is_empty(self):
        kvf = KVFile(self.temp_dir + '/foo.yml')
        default = object()
        assert kvf.get('foo', default) is default

    def test_load_existent(self):
        file_name = self.temp_dir + '/foo.yml'
        with open(file_name, 'w') as fp:
            fp.write('foo: bar')

        assert KVFile(file_name).get('foo', '') == 'bar'

    def test_load_non_dictionary(self):
        file_name = self.temp_dir + '/foo.yml'
        with open(file_name, 'w') as fp:
            fp.write('[not, a, dictionary]')

        with pytest.raises(TypeError) as excinfo:
            KVFile(file_name)
        assert 'not a dictionary' in six.text_type(excinfo.value)

    def test_get_after_set_same(self):
        kvf = KVFile(self.temp_dir + '/foo.yml')
        kvf.set_and_save('foo', 'bar')

        assert kvf.get('foo', '') == 'bar'

    def test_load_second_same(self):
        kvf = KVFile(self.temp_dir + '/foo.yml')
        kvf.set_and_save('foo', 'bar')
        kvf2 = KVFile(self.temp_dir + '/foo.yml')

        assert kvf2.get('foo', '') == 'bar'

    def test_sets_dont_cause_append_duplication(self):
        file_name = self.temp_dir + '/foo.yml'
        kvf = KVFile(file_name)
        kvf.set_and_save('foo', 'bar')
        kvf.set_and_save('foo2', 'bar')

        with open(file_name, 'r') as fp:
            lines = fp.readlines()
            fp.seek(0)
            data = yaml.safe_load(fp)

        assert len(lines) == 2
        assert data == {'foo': 'bar', 'foo2': 'bar'}



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


from django.test import SimpleTestCase

from django_perf_rec.utils import current_test, sorted_names


class CurrentTestTests(SimpleTestCase):

    def test_here(self):
        details = current_test()
        assert details.file_path == __file__
        assert details.class_name == 'CurrentTestTests'
        assert details.test_name == 'test_here'

    def test_twice_same(self):
        assert current_test() == current_test()

    def test_functional(self):

        def test_thats_functional():
            return current_test()

        details = test_thats_functional()
        assert details.file_path == __file__
        assert details.class_name is None
        assert details.test_name == 'test_thats_functional'


class SortedNamesTests(SimpleTestCase):

    def test_empty(self):
        assert sorted_names([]) == []

    def test_just_default(self):
        assert sorted_names(['default']) == ['default']

    def test_just_something(self):
        assert sorted_names(['something']) == ['something']

    def test_does_sort(self):
        assert sorted_names(['b', 'a']) == ['a', 'b']

    def test_sort_keeps_default_first(self):
        assert sorted_names(['a', 'default']) == ['default', 'a']



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from django_perf_rec.sql import sql_fingerprint


def test_select():
    assert (
        sql_fingerprint("SELECT `f1`, `f2` FROM `b`") ==
        "SELECT ... FROM `b`"
    )


def test_select_where():
    assert (
        sql_fingerprint("SELECT DISTINCT `table`.`field` FROM `table` WHERE `table`.`id` = 1") ==
        "SELECT DISTINCT `table`.`field` FROM `table` WHERE `table`.`id` = #"
    )


def test_select_comment():
    assert (
        sql_fingerprint("SELECT /* comment */ `f1`, `f2` FROM `b`") ==
        "SELECT /* comment */ ... FROM `b`"
    )


def test_select_join():
    assert (
        sql_fingerprint('SELECT f1, f2 FROM a INNER JOIN b ON (a.b_id = b.id) WHERE a.f2 = 1') ==
        'SELECT ... FROM a INNER JOIN b ON (a.b_id = b.id) WHERE a.f2 = #'
    )


def test_select_order_by():
    assert (
        sql_fingerprint('SELECT f1, f2 FROM a ORDER BY f3, f4') ==
        'SELECT ... FROM a ORDER BY ...'
    )


def test_insert():
    assert (
        sql_fingerprint("INSERT INTO `table` (`f1`, `f2`) VALUES ('v1', 2)") ==
        "INSERT INTO `table` (...) VALUES (...)"
    )


def test_update():
    assert (
        sql_fingerprint("UPDATE `table` SET `foo` = 'bar' WHERE `table`.`id` = 1") ==
        "UPDATE `table` SET ... WHERE `table`.`id` = #"
    )


def test_savepoint():
    assert (
        sql_fingerprint("SAVEPOINT `s140323809662784_x54`") ==
        "SAVEPOINT `#`"
    )


def test_rollback_to_savepoint():
    assert (
        sql_fingerprint("ROLLBACK TO SAVEPOINT `s140323809662784_x54`") ==
        "ROLLBACK TO SAVEPOINT `#`"
    )


def test_release_savepoint():
    assert (
        sql_fingerprint("RELEASE SAVEPOINT `s140699855320896_x17`") ==
        "RELEASE SAVEPOINT `#`"
    )



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import mock
import pytest
from django.core.cache import caches
from django.test import SimpleTestCase, TestCase

from django_perf_rec.cache import AllCacheRecorder, CacheOp, CacheRecorder


class CacheOpTests(SimpleTestCase):

    def test_clean_key_integer(self):
        assert CacheOp.clean_key('foo1') == 'foo#'

    def test_clean_key_uuid(self):
        assert CacheOp.clean_key('bdfc9986-d461-4a5e-bf98-8688993abcfb') == '#'

    def test_clean_key_random_hash(self):
        assert CacheOp.clean_key('abc123abc123abc123abc123abc12345') == '#'

    def test_key(self):
        op = CacheOp('default', 'foo', 'bar')
        assert op.alias == 'default'
        assert op.operation == 'foo'
        assert op.key_or_keys == 'bar'

    def test_keys(self):
        op = CacheOp('default', 'foo', ['bar', 'baz'])
        assert op.alias == 'default'
        assert op.operation == 'foo'
        assert op.key_or_keys == ['bar', 'baz']

    def test_invalid(self):
        with pytest.raises(ValueError):
            CacheOp('x', 'foo', object())

    def test_equal(self):
        assert (
            CacheOp('x', 'foo', 'bar') ==
            CacheOp('x', 'foo', 'bar')
        )

    def test_not_equal_alias(self):
        assert (
            CacheOp('x', 'foo', 'bar') !=
            CacheOp('y', 'foo', 'bar')
        )

    def test_not_equal_operation(self):
        assert (
            CacheOp('x', 'foo', 'bar') !=
            CacheOp('x', 'bar', 'bar')
        )

    def test_not_equal_keys(self):
        assert (
            CacheOp('x', 'foo', ['bar']) !=
            CacheOp('x', 'foo', ['baz'])
        )


class CacheRecorderTests(TestCase):

    def test_default(self):
        callback = mock.Mock()
        with CacheRecorder('default', callback):
            caches['default'].get('foo')
        callback.assert_called_once_with(
            CacheOp('default', 'get', 'foo')
        )

    def test_secondary(self):
        callback = mock.Mock()
        with CacheRecorder('second', callback):
            caches['second'].get('foo')
        callback.assert_called_once_with(
            CacheOp('second', 'get', 'foo')
        )

    def test_secondary_default_not_recorded(self):
        callback = mock.Mock()
        with CacheRecorder('second', callback):
            caches['default'].get('foo')
        assert len(callback.mock_calls) == 0


class AllCacheRecorderTests(TestCase):

    def test_records_all(self):
        callback = mock.Mock()
        with AllCacheRecorder(callback):
            caches['default'].get('foo')
            caches['second'].set('bar', 'baz')
            caches['default'].delete_many(['foo'])

        assert callback.mock_calls == [
            mock.call(CacheOp('default', 'get', 'foo')),
            mock.call(CacheOp('second', 'set', 'bar')),
            mock.call(CacheOp('default', 'delete_many', ['foo'])),
        ]



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from django.test import SimpleTestCase

from django_perf_rec.orm import patch_ORM_to_be_deterministic


class PatchORMToBeDeterministicTests(SimpleTestCase):

    def test_call_it(self):
        patch_ORM_to_be_deterministic()

    def test_call_it_again(self):
        patch_ORM_to_be_deterministic()



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import errno
import os

import pytest
import six
from django.core.cache import caches
from django.db.models import Q
from django.db.models.functions import Upper
from django.test import TestCase

from django_perf_rec import TestCaseMixin, record
from testapp.models import Author

from .utils import run_query


class RecordTests(TestCase):

    def test_single_db_query(self):
        with record():
            run_query('default', 'SELECT 1337')

    def test_multiple_db_queries(self):
        with record():
            run_query('default', 'SELECT 1337')
            run_query('default', 'SELECT 4949')

    def test_non_deterministic_QuerySet_annotate(self):
        with record():
            list(Author.objects.annotate(
                x=Upper('name'),
                y=Upper('name'),
            ))

    def test_non_deterministic_QuerySet_extra(self):
        with record():
            list(Author.objects.extra(select={
                'x': '1',
                'y': '1',
            }))

    def test_non_deterministic_Q_query(self):
        with record():
            list(Author.objects.filter(Q(name='foo', age=1)))

    def test_single_cache_op(self):
        with record():
            caches['default'].get('foo')

    def test_multiple_cache_ops(self):
        with record():
            caches['default'].set('foo', 'bar')
            caches['second'].get_many(['foo', 'bar'])
            caches['default'].delete('foo')

    def test_multiple_calls_in_same_function_are_different_records(self):
        with record():
            caches['default'].get('foo')

        with record():
            caches['default'].get('bar')

    def test_custom_name(self):
        with record(record_name='custom'):
            caches['default'].get('foo')

    def test_custom_name_multiple_calls(self):
        with record(record_name='custom'):
            caches['default'].get('foo')

        with pytest.raises(AssertionError) as excinfo:
            with record(record_name='custom'):
                caches['default'].get('bar')

        assert 'Performance record did not match' in six.text_type(excinfo.value)


class RecordBlankFileTests(TestCase):

    perf_name = __file__.replace('.py', '.empty.perf.yml')

    @classmethod
    def setUpClass(cls):
        super(RecordBlankFileTests, cls).setUpClass()
        cls.ensure_no_file()

    @classmethod
    def tearDownClass(cls):
        cls.ensure_no_file()
        super(RecordBlankFileTests, cls).tearDownClass()

    @classmethod
    def ensure_no_file(cls):
        try:
            os.unlink(cls.perf_name)
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise

    def test_one(self):
        with record(file_name=self.perf_name):
            caches['default'].get('foo')

    def test_two(self):
        with record(file_name=self.perf_name):
            caches['default'].get('foo')
        with record(file_name=self.perf_name):
            caches['default'].get('foo')


class TestCaseMixinTests(TestCaseMixin, TestCase):

    def test_record_performance(self):
        with self.record_performance():
            caches['default'].get('foo')

    def test_record_performance_record_name(self):
        with self.record_performance(record_name='other'):
            caches['default'].get('foo')

    def test_record_performance_file_name(self):
        perf_name = __file__.replace('.py', '.file_name.perf.yml')
        with self.record_performance(file_name=perf_name):
            caches['default'].get('foo')



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import inspect
from collections import namedtuple


TestDetails = namedtuple('TestDetails', ['file_path', 'class_name', 'test_name'])


def current_test():
    """
    Use a little harmless stack inspection to determine the test that is currently running.
    """
    frame = inspect.currentframe()
    try:
        while True:
            if frame.f_code.co_name.startswith('test_'):
                break

            # Next frame
            frame = frame.f_back
            if frame is None:
                break

        if frame is None:
            raise RuntimeError("Could not automatically determine the test name.")

        file_path = frame.f_globals['__file__']

        # May be a pytest function test so we can't assume 'self' exists
        its_self = frame.f_locals.get('self', None)
        if its_self is None:
            class_name = None
        else:
            class_name = its_self.__class__.__name__

        test_name = frame.f_code.co_name

        return TestDetails(
            file_path=file_path,
            class_name=class_name,
            test_name=test_name,
        )
    finally:
        # Always delete frame references to help garbage collector
        del frame


def sorted_names(names):
    """
    Sort a list of names but keep the word 'default' first if it's there.
    """
    names = list(names)

    have_default = False
    if 'default' in names:
        names.remove('default')
        have_default = True

    sorted_names = sorted(names)

    if have_default:
        sorted_names = ['default'] + sorted_names

    return sorted_names



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from threading import local

from django.core.cache import DEFAULT_CACHE_ALIAS
from django.db import DEFAULT_DB_ALIAS

from .cache import AllCacheRecorder
from .db import AllDBRecorder
from .utils import current_test
from .yaml import KVFile


record_current = local()


def record(file_name=None, record_name=None):
    test_details = current_test()

    if file_name is None:
        file_name = test_details.file_path
        if file_name.endswith('.py'):
            file_name = file_name[:-len('.py')] + '.perf.yml'
        else:
            file_name += '.perf.yml'

    if record_name is None:
        if test_details.class_name:
            record_name = '{class_}.{test}'.format(
                class_=test_details.class_name,
                test=test_details.test_name,
            )
        else:
            record_name = test_details.test_name

        # Multiple calls inside the same test should end up suffixing with .2, .3 etc.
        if getattr(record_current, 'record_name', None) == record_name:
            record_current.counter += 1
            record_name = record_name + '.{}'.format(record_current.counter)
        else:
            record_current.record_name = record_name
            record_current.counter = 1

    return PerformanceRecorder(file_name, record_name)


class PerformanceRecorder(object):

    def __init__(self, file_name, record_name):
        self.file_name = file_name
        self.record_name = record_name

        self.record = []
        self.db_recorder = AllDBRecorder(self.on_db_op)
        self.cache_recorder = AllCacheRecorder(self.on_cache_op)

    def __enter__(self):
        self.db_recorder.__enter__()
        self.cache_recorder.__enter__()
        self.load_recordings()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.cache_recorder.__exit__(exc_type, exc_value, exc_traceback)
        self.db_recorder.__exit__(exc_type, exc_value, exc_traceback)

        if exc_type is None:
            self.save_or_assert()

    def on_db_op(self, db_op):
        name_parts = ['db']
        if db_op.alias != DEFAULT_DB_ALIAS:
            name_parts.append(db_op.alias)
        name = '|'.join(name_parts)

        self.record.append({name: db_op.sql})

    def on_cache_op(self, cache_op):
        name_parts = ['cache']
        if cache_op.alias != DEFAULT_CACHE_ALIAS:
            name_parts.append(cache_op.alias)
        name_parts.append(cache_op.operation)
        name = '|'.join(name_parts)

        self.record.append({name: cache_op.key_or_keys})

    def load_recordings(self):
        self.records_file = KVFile(self.file_name)

    def save_or_assert(self):
        orig_record = self.records_file.get(self.record_name, None)

        if orig_record is not None:
            assert self.record == orig_record, "Performance record did not match for {}".format(self.record_name)

        self.records_file.set_and_save(self.record_name, self.record)


class TestCaseMixin(object):
    """
    Adds record_performance() method to TestCase class it's mixed into
    for easy import-free use.
    """
    def record_performance(self, file_name=None, record_name=None):
        return record(file_name=file_name, record_name=record_name)



# -*- coding:utf-8 -*-
"""
isort:skip_file
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import six

try:
    import pytest
except ImportError:
    pytest = None

if pytest is not None:
    if six.PY2:
        pytest.register_assert_rewrite(b'django_perf_rec.api')
    else:
        pytest.register_assert_rewrite('django_perf_rec.api')

from .api import TestCaseMixin, record  # noqa: F401

__version__ = '1.0.2'



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from django.db.backends.sqlite3.base import DatabaseWrapper as OrigDatabaseWrapper

from .operations import DatabaseOperations


class DatabaseWrapper(OrigDatabaseWrapper):

    def __init__(self, *args, **kwargs):
        super(DatabaseWrapper, self).__init__(*args, **kwargs)
        self.ops = DatabaseOperations(self)



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os

import django

BASE_DIR = os.path.dirname(os.path.dirname(__file__))

DEBUG = True
TEMPLATE_DEBUG = DEBUG

SECRET_KEY = 'NOTASECRET'


if django.VERSION[:2] == (1, 8):
    # Use a wrapper that includes Django commit 4f6a7663bcddffb114f2647f9928cbf1fdd8e4b5
    # so that full SQL queries from sqlite come through
    engine = 'django18_sqlite3_backend'
else:
    engine = 'django.db.backends.sqlite3'

DATABASES = {
    'default': {
        'ENGINE': engine,
        'NAME': ':memory:',
    },
    'replica': {
        'ENGINE': engine,
        'NAME': ':memory:',
        'TEST': {
            'MIRROR': 'default',
        }
    },
    'second': {
        'ENGINE': engine,
        'NAME': ':memory:',
    },
}

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
    },
    'second': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
    },
}

ALLOWED_HOSTS = []

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'testapp',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'urls'
LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import six
from django.utils.lru_cache import lru_cache
from sqlparse import parse, tokens
from sqlparse.sql import IdentifierList, Token


@lru_cache(maxsize=500)
def sql_fingerprint(query):
    """
    Simplify a query, taking away exact values and fields selected.

    Imperfect but better than super explicit, value-dependent queries.
    """
    parsed_query = parse(query)[0]
    sql_recursively_simplify(parsed_query)
    return six.text_type(parsed_query)


sql_deleteable_tokens = (
    tokens.Number,
    tokens.Number.Float,
    tokens.Number.Integer,
    tokens.Number.Hexadecimal,
    tokens.String,
    tokens.String.Single,
)


def sql_recursively_simplify(node):
    # Erase which fields are being updated in an UPDATE
    if node.tokens[0].value == 'UPDATE':
        i_set = [i for (i, t) in enumerate(node.tokens) if t.value == 'SET'][0]
        i_where = [i for (i, t) in enumerate(node.tokens)
                   if t.is_group() and t.tokens[0].value == 'WHERE'][0]
        middle = [Token(tokens.Punctuation, ' ... ')]
        node.tokens = node.tokens[:i_set + 1] + middle + node.tokens[i_where:]

    # Erase the names of savepoints since they are non-deteriministic
    if hasattr(node, 'tokens'):
        # SAVEPOINT x
        if str(node.tokens[0]) == 'SAVEPOINT':
            node.tokens[2].tokens[0].value = '`#`'
            return
        # RELEASE SAVEPOINT x
        elif len(node.tokens) >= 3 and node.tokens[2].value == 'SAVEPOINT':
            node.tokens[4].tokens[0].value = "`#`"
            return
        # ROLLBACK TO SAVEPOINT X
        token_values = [getattr(t, 'value', '') for t in node.tokens]
        if len(node.tokens) == 7 and token_values[:6] == ['ROLLBACK', ' ', 'TO', ' ', 'SAVEPOINT', ' ']:
            node.tokens[6].tokens[0].value = '`#`'
            return

    for i, token in enumerate(node.tokens):
        ttype = getattr(token, 'ttype', None)

        if isinstance(token, IdentifierList):
            token.tokens = [Token(tokens.Punctuation, '...')]
        elif hasattr(token, 'tokens'):
            sql_recursively_simplify(token)
        elif ttype in sql_deleteable_tokens:
            token.value = '#'
        elif ttype == tokens.Whitespace.Newline:
            token.value = ''  # Erase newlines
        elif ttype == tokens.Whitespace:
            token.value = ' '
        elif getattr(token, 'value', None) == 'NULL':
            token.value = '#'



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from django.db.backends.sqlite3.operations import DatabaseOperations as OrigDatabaseOperations


class DatabaseOperations(OrigDatabaseOperations):

    # From Django commit 4f6a7663bcddffb114f2647f9928cbf1fdd8e4b5

    def _quote_params_for_last_executed_query(self, params):
        """
        Only for last_executed_query! Don't use this to execute SQL queries!
        """
        sql = 'SELECT ' + ', '.join(['QUOTE(?)'] * len(params))
        # Bypass Django's wrappers and use the underlying sqlite3 connection
        # to avoid logging this query - it would trigger infinite recursion.
        cursor = self.connection.connection.cursor()
        # Native sqlite3 cursors cannot be used as context managers.
        try:
            return cursor.execute(sql, params).fetchone()
        finally:
            cursor.close()

    def last_executed_query(self, cursor, sql, params):
        # Python substitutes parameters in Modules/_sqlite/cursor.c with:
        # pysqlite_statement_bind_parameters(self->statement, parameters, allow_8bit_chars);
        # Unfortunately there is no way to reach self->statement from Python,
        # so we quote and substitute parameters manually.
        if params:
            if isinstance(params, (list, tuple)):
                params = self._quote_params_for_last_executed_query(params)
            else:
                keys = params.keys()
                values = tuple(params.values())
                values = self._quote_params_for_last_executed_query(values)
                params = dict(zip(keys, values))
            return sql % params
        # For consistency with SQLiteCursorWrapper.execute(), just return sql
        # when there are no parameters. See #13648 and #17158.
        else:
            return sql



#!/usr/bin/env python
# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import re
import sys

from setuptools import find_packages, setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


version = get_version('django_perf_rec')


if sys.argv[-1] == 'publish':
    if os.system("pip freeze | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload dist/*")
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    sys.exit()


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

setup(
    name='django-perf-rec',
    version=version,
    description="Keep detailed records of the performance of your Django "
                "code.",
    long_description=readme + '\n\n' + history,
    author='YPlan',
    author_email='adam@yplanapp.com',
    url='https://github.com/YPlan/django-perf-rec',
    packages=find_packages(exclude=['test', 'test.*']),
    include_package_data=True,
    install_requires=[
        'Django',
        'patchy',
        'PyYAML',
        'six',
        'sqlparse>=0.2.0',
    ],
    license='MIT',
    zip_safe=False,
    keywords='Django',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import mock
from django.test import SimpleTestCase, TestCase

from django_perf_rec.db import AllDBRecorder, DBOp, DBRecorder

from .utils import run_query


class DBOpTests(SimpleTestCase):

    def test_create(self):
        op = DBOp('myalias', 'SELECT 1')
        assert op.alias == 'myalias'
        assert op.sql == 'SELECT 1'

    def test_equal(self):
        assert (
            DBOp('foo', 'bar') ==
            DBOp('foo', 'bar')
        )

    def test_not_equal_alias(self):
        assert (
            DBOp('foo', 'bar') !=
            DBOp('baz', 'bar')
        )

    def test_not_equal_sql(self):
        assert (
            DBOp('foo', 'bar') !=
            DBOp('foo', 'baz')
        )


class DBRecorderTests(TestCase):

    def test_default(self):
        callback = mock.Mock()
        with DBRecorder('default', callback):
            run_query('default', 'SELECT 1')
        callback.assert_called_once_with(
            DBOp('default', 'SELECT #')
        )

    def test_secondary(self):
        callback = mock.Mock()
        with DBRecorder('second', callback):
            run_query('second', 'SELECT 1')
        callback.assert_called_once_with(
            DBOp('second', 'SELECT #')
        )

    def test_replica(self):
        callback = mock.Mock()
        with DBRecorder('replica', callback):
            run_query('replica', 'SELECT 1')
        callback.assert_called_once_with(
            DBOp('replica', 'SELECT #')
        )

    def test_secondary_default_not_recorded(self):
        callback = mock.Mock()
        with DBRecorder('second', callback):
            run_query('default', 'SELECT 1')
        assert len(callback.mock_calls) == 0


class AllDBRecorderTests(TestCase):

    def test_records_all(self):
        callback = mock.Mock()
        with AllDBRecorder(callback):
            run_query('replica', 'SELECT 1')
            run_query('default', 'SELECT 2')
            run_query('second', 'SELECT 3')

        assert callback.mock_calls == [
            mock.call(DBOp('replica', 'SELECT #')),
            mock.call(DBOp('default', 'SELECT #')),
            mock.call(DBOp('second', 'SELECT #')),
        ]
